# Copyright (C) 2017  Lars Wirzenius
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import ick2


class WorkAPI(ick2.APIbase):

    def __init__(self, state):
        super().__init__(state)
        self._workers = Workers(state)
        self._projects = Projects(state)
        self._type_name = 'work'

    def get_routes(self, path):  # pragma: no cover
        return [
            {
                'method': 'GET',
                'path': '{}/<worker>'.format(path),
                'callback': self.GET(self.get_work),
            },
            {
                'method': 'POST',
                'path': path,
                'callback': self.POST(self.update_work),
            },
        ]

    def get_work(self, worker, **kwargs):
        worker_state = self._workers.get_worker(worker)
        if not worker_state.get('doing'):
            project, pipeline = self._pick_triggered_pipeline()
            if project is None:
                doing = {}
            else:
                pipeline['status'] = 'building'
                self._update_pipeline(project, pipeline)

                build_id = project.get('build_id', 0) + 1
                project['build_id'] = build_id
                self._projects.update_project(project)

                self._start_build(project, pipeline, worker, build_id)
                self._start_log(build_id)

                index = 0
                doing = {
                    'build_id': build_id,
                    'worker': worker,
                    'project': project['project'],
                    'pipeline': pipeline['name'],
                    'parameters': project.get('parameters', {}),
                    'fresh_workspace': True,
                    'step': pipeline['actions'][index],
                    'step_index': index,
                    'log': '/logs/{}'.format(build_id),
                }

            worker_state = {
                'worker': worker,
                'doing': doing,
            }
            self._workers.update_worker(worker_state)

        return worker_state['doing']

    def _pick_triggered_pipeline(self):
        projects = self._projects.get_projects()
        for project in projects:
            for name in project['pipelines']:
                pp = self._pipeline_instance_name(project['project'], name)
                try:
                    pl = self._state.get_resource('pipeline_instances', pp)
                except ick2.NotFound:
                    pass
                else:
                    if pl.get('status') == 'triggered':
                        pt = self._state.get_resource('pipelines', name)
                        return project, pt
        return None, None

    def _pipeline_instance_name(self, project_name, pipeline_name):
        return '{} {}'.format(project_name, pipeline_name)

    def _update_pipeline(self, project, pipeline):
        pp = self._pipeline_instance_name(project['project'], pipeline['name'])
        try:
            self._state.get_resource('pipeline_instances', pp)
        except ick2.NotFound:  # pragma: no cover
            self._state.add_resource('pipeline_instances', pp, pipeline)
        else:
            self._state.update_resource('pipeline_instances', pp, pipeline)

    def _start_build(self, project, pipeline, worker, build_id):
        ick2.log.log('info', msg_text='Starting new build', build_id=build_id)
        build = {
            'build_id': build_id,
            'log': '/logs/{}'.format(build_id),
            'worker': worker,
            'project': project['project'],
            'pipeline': pipeline['name'],
            'status': 'building',
        }
        self._state.add_resource('builds', str(build_id), build)
        return build_id

    def _start_log(self, build_id):
        ick2.log.log('info', msg_text='Starting new log', build_id=build_id)
        log = {
            'build_id': build_id,
            'log': '',
        }
        self._state.add_resource('log', str(build_id), log)

    def update_work(self, update, **kwargs):
        if 'worker' not in update:  # pragma: no cover
            raise ick2.BadUpdate('no worker specified')

        worker_state = self._workers.get_worker(update['worker'])
        doing = worker_state.get('doing', {})
        self._check_work_update(doing, update)

        project, pipeline = self._get_pipeline(
            update['project'], update['pipeline'])
        self._append_to_build_log(update)

        ick2.log.log(
            'trace',
            msg_text='xxx update_work',
            update=update,
            project=project,
            pipeline=pipeline,
            doing=doing)

        exit_code = update.get('exit_code')
        if exit_code == 0:
            index = doing['step_index'] + 1
            actions = pipeline['actions']
            if index >= len(actions):
                pipeline['status'] = 'idle'
                self._update_pipeline(project, pipeline)
                doing = {}
                self._finish_build(update)
            else:
                doing['step_index'] = index
                doing['step'] = dict(actions[index])
                doing['fresh_workspace'] = False

            worker_state = {
                'worker': update['worker'],
                'doing': doing,
            }
            self._workers.update_worker(worker_state)
        elif exit_code is not None:
            assert isinstance(exit_code, int)
            assert exit_code != 0
            pipeline['status'] = 'idle'
            self._update_pipeline(project, pipeline)
            self._finish_build(update)

            worker_state = {
                'worker': update['worker'],
                'doing': {},
            }
            self._workers.update_worker(worker_state)

    def _check_work_update(self, doing, update):  # pragma: no cover
        must_match = ['worker', 'project', 'pipeline', 'build_id']
        for name in must_match:
            if name not in update:
                raise ick2.BadUpdate('{} not specified'.format(name))
            if doing.get(name) != update[name]:
                raise ick2.BadUpdate(
                    '{} differs from current work: {} vs {}'.format(
                        name, doing.get(name), update[name]))

    def _get_pipeline(self, project, pipeline):  # pragma: no cover
        projects = [
            p
            for p in self._projects.get_projects()
            if p['project'] == project
        ]
        if not projects:
            raise ick2.NotFound()
        p = projects[0]
        if pipeline not in p['pipelines']:
            raise ick2.NotFound()
        pt = self._state.get_resource('pipelines', pipeline)
        return p, pt

    def _append_to_build_log(self, update):
        build_id = update['build_id']
        log = self._state.get_resource('log', str(build_id))
        for kind in ['stdout', 'stderr']:
            text = update.get(kind, '')
            if text is not None:
                log['log'] += text
        self._state.update_resource('log', str(build_id), log)

    def _finish_build(self, update):
        build = self._state.get_resource('builds', str(update['build_id']))
        build['status'] = update['exit_code']
        self._state.update_resource('builds', str(update['build_id']), build)

    def create(self, *args, **kwargs):  # pragma: no cover
        pass

    def update(self, *args, **kwargs):  # pragma: no cover
        pass

    def list(self, *args, **kwargs):  # pragma: no cover
        pass

    def show(self, *args, **kwargs):  # pragma: no cover
        pass

    def delete(self, *args, **kwargs):  # pragma: no cover
        pass


class Workers:  # pragma: no cover

    def __init__(self, state):
        self._state = state

    def get_worker(self, name):
        try:
            return self._state.get_resource('workers', name)
        except ick2.NotFound:
            return {
                'worker': name,
            }

    def update_worker(self, worker):
        try:
            self._state.get_resource('workers', worker['worker'])
        except ick2.NotFound:
            raise ick2.NotFound()
        else:
            self._state.update_resource(
                'workers', worker['worker'], worker)


class Projects:  # pragma: no cover

    def __init__(self, state):
        self._state = state

    def get_projects(self):
        return self._state.get_resources('projects')

    def update_project(self, project):
        self._state.update_resource('projects', project['project'], project)
